#! /usr/bin/env python3

"""Post-install / configuration script for tapeimgr"""

import os
import io
import json
import sys
import argparse
from shutil import copyfile


def parseCommandLine(parser):
    """Parse command line"""

    parser.add_argument('--remove', '-r',
                        action='store_true',
                        dest='removeFlag',
                        default=False,
                        help='remove all tapeimgr configuration files')
    # Parse arguments
    args = parser.parse_args()
    return args


def errorExit(msg):
    """Print error to stderr and exit"""
    msgString = ('ERROR: ' + msg + '\n')
    sys.stderr.write(msgString)
    sys.exit(1)

def infoMessage(msg):
    """Print message to stderr"""
    msgString = ('INFO: ' + msg + '\n')
    sys.stderr.write(msgString)

def writeConfigFile(sudoUser, sudoUID, sudoGID, removeFlag):
    """Create configuration file"""

    # Exit if script nor tun as root
    if sudoUID is None or sudoGID is None:
        msg = 'this script must be run as root'
        errorExit(msg)

    # Create configuration directory under /etc/
    etcDir = os.path.normpath('/etc/')

    if not os.access(etcDir, os.W_OK | os.X_OK):
        msg = 'cannot write to ' + etcDir
        errorExit(msg)

    configDir = os.path.join(etcDir, 'tapeimgr')

    if not removeFlag:
        if not os.path.isdir(configDir):
            os.mkdir(configDir)

    # Path to configuration file
    fConfig = os.path.join(configDir, 'tapeimgr.json')

    # Dictionary with items in configuration file
    configSettings = {}
    configSettings['SUDO_USER'] = sudoUser
    configSettings['SUDO_UID'] = sudoUID
    configSettings['SUDO_GID'] = sudoGID
    configSettings['files'] = ''
    configSettings['logFileName'] = 'tapeimgr.log'
    configSettings['tapeDevice'] = '/dev/nst0'
    configSettings['initBlockSize'] = '512'
    configSettings['prefix'] = 'file'
    configSettings['extension'] = 'dd'
    configSettings['fillBlocks'] = 'False'

    if not removeFlag:
        # Write to configuration file in json format
        infoMessage('writing configuration file ' + fConfig)
        with io.open(fConfig, 'w', encoding='utf-8') as f:
            json.dump(configSettings, f, indent=4, sort_keys=True)
    else:
        if os.path.isfile(fConfig):
            infoMessage('removing configuration file ' + fConfig)
            os.remove(fConfig)
        if os.path.isdir(configDir):
            infoMessage('removing configuration directory ' + configDir)
            os.rmdir(configDir)

def writeDesktopFiles(packageDir, sudoUID, sudoGID, removeFlag):
    """Creates desktop files in /usr/share/applications and on desktop"""

    # Locate icon file in package
    iconFile = os.path.join(packageDir, 'icons', 'tapeimgr.png')
    if not os.path.isfile(iconFile):
        msg = 'cannot find icon file'
        errorExit(msg)

    # Locate /etc, applications and desktop directories
    desktopDir = os.path.join(os.path.join(os.path.expanduser('~')), 'Desktop')
    applicationsDir = os.path.normpath('/usr/share/applications')

    # Check if directories are writable
    if not os.access(desktopDir, os.W_OK | os.X_OK):
        msg = 'cannot write to ' + desktopDir
        errorExit(msg)

    if not os.access(applicationsDir, os.W_OK | os.X_OK):
        msg = 'cannot write to ' + applicationsDir
        errorExit(msg)

    fDesktop = os.path.join(desktopDir, 'tapeimgr.desktop')
    fApplications = os.path.join(applicationsDir, 'tapeimgr.desktop')

    # List of desktop file lines
    desktopList = []
    desktopList.append('[Desktop Entry]')
    desktopList.append('Type=Application')
    desktopList.append('Encoding=UTF-8')
    desktopList.append('Name=tapeimgr')
    desktopList.append('Comment=Simple tape imaging and extraction tool')
    desktopList.append('Exec=tapeimgr-pkexec')
    desktopList.append('Icon=' + iconFile)
    desktopList.append('Terminal=false')
    desktopList.append('Categories=Utility;System;GTK')

    # Write desktop file to Desktop
    if not removeFlag:
        try:
            infoMessage('creating desktop file ' + fDesktop)
            with io.open(fDesktop, 'w', encoding='utf-8') as fD:
                for line in desktopList:
                    fD.write(line + '\n')
            # Change owner to user (since script is executed as root)
            os.chown(fDesktop, int(sudoUID), int(sudoGID))
        except:
            msg = 'Failed to create ' + fDesktop
            errorExit(msg)

        # Write desktop file to applications directory
        try:
            infoMessage('creating desktop file ' + fApplications)
            with io.open(fApplications, 'w', encoding='utf-8') as fA:
                for line in desktopList:
                    fA.write(line + '\n')
        except:
            msg = 'Failed to create ' + fApplications
            errorExit(msg)
    else:
        if os.path.isfile(fDesktop):
            infoMessage('removing desktop file ' + fDesktop)
            os.remove(fDesktop)
        if os.path.isfile(fApplications):
            infoMessage('removing desktop file ' + fApplications)
            os.remove(fApplications)


def  writePKPolicyFile(packageDir, removeFlag):
    """Creates policy file in /usr/share/polkit-1/actions which is required to
    launch tapeimgr with pkexec"""

    policyFileName = 'com.ubuntu.pkexec.tapeimgr.policy'

    policyFileIn = os.path.join(packageDir, 'pkexec', policyFileName)
    if not os.path.isfile(policyFileIn):
        msg = 'cannot find policy file'
        errorExit(msg)

    # Locate polkit actions dir and check if we can write there
    actionsDir = os.path.normpath('/usr/share/polkit-1/actions')
    if not os.path.isdir(actionsDir):
        msg = 'cannot find actions dir'
        errorExit(msg)

    # Construct path to output policy file
    policyFileOut = os.path.join(actionsDir, policyFileName)

    # Copy policy file to actions dir
    if not removeFlag:
        try:
            infoMessage('creating policy file ' + policyFileOut)
            copyfile(policyFileIn, policyFileOut)
        except IOError:
            msg = 'could not copy policy file to ' + policyFileOut
            errorExit(msg)
    else:
        if os.path.isfile(policyFileOut):
            infoMessage('removing policy file ' + policyFileOut)
            os.remove(policyFileOut)

def writePKLauncher(packageDir, removeFlag):
    """Creates launcher script in /usr/local/bin which is use by pkexec to
    launch tapeimgr"""

    pkExecLauncherFileName = 'tapeimgr-pkexec'

    pkExecLauncherIn = os.path.join(packageDir, 'pkexec', pkExecLauncherFileName)

    if not os.path.isfile(pkExecLauncherIn):
        msg = 'cannot find pkExec launcher file'
        errorExit(msg)

    # Locate usr/local/bin dir and check if we can write there
    binDir = os.path.normpath('/usr/local/bin')
    if not os.path.isdir(binDir):
        msg = 'cannot find ' + binDir
        errorExit(msg)

    if not os.access(binDir, os.W_OK | os.X_OK):
        msg = 'cannot write to ' + binDir
        errorExit(msg)

    # Construct path to output pk launcher
    pkExecLauncherOut = os.path.join(binDir, pkExecLauncherFileName)

    # Copy pk launcher to bin dir and make it executable
    if not removeFlag:
        try:
            infoMessage('creating pkexec launcher ' + pkExecLauncherOut)
            copyfile(pkExecLauncherIn, pkExecLauncherOut)
            os.chmod(pkExecLauncherOut, 0o755)
        except IOError:
            msg = 'could not copy pkexec launcher to ' + pkExecLauncherOut
            errorExit(msg)
    else:
        if os.path.isfile(pkExecLauncherOut):
            infoMessage('removing pkexec launcher file ' + pkExecLauncherOut)
            os.remove(pkExecLauncherOut)


def main():
    """
    Creates the following items:
    - configuration directory tapeimgr in /etc/
    - configuration file in /etc/tapeimgr
    - pkexec policy file in /usr/share/polkit-1/actions
    - pkexec launcher script in /usr/local/bin
    - desktop file in /usr/share/applications
    - desktop file on user's desktop
    If the --remove / -r switch is given the above items
    are removed (if they exist)
    """

    # Parse command line
    parser = argparse.ArgumentParser(description='tapeingr configuration tool')
    args = parseCommandLine(parser)
    removeFlag = args.removeFlag

    # Package dir
    packageDir = os.path.dirname(os.path.abspath(__file__))

    # Get evironment variables
    sudoUser = os.environ.get('SUDO_USER')
    sudoUID = os.environ.get('SUDO_UID')
    sudoGID = os.environ.get('SUDO_GID')

    writeConfigFile(sudoUser, sudoUID, sudoGID, removeFlag)
    writePKLauncher(packageDir, removeFlag)
    writePKPolicyFile(packageDir, removeFlag)
    writeDesktopFiles(packageDir, sudoUID, sudoGID, removeFlag)
    infoMessage('tapeimgr configuration completed successfully!')


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
"""
Tapeimgr, automated reading of tapes

Author: Johan van der Knijff
Research department,  KB / National Library of the Netherlands
"""
import sys
from .cli import main as cliLaunch
from .gui import main as guiLaunch
from . import config

__version__ = '0.2.0'

def main():
    """Launch GUI if no command line arguments were given; otherwise launch CLI"""
    config.version = __version__
    noArgs = len(sys.argv)
    if noArgs == 1:
        guiLaunch()
    else:
        cliLaunch()

main()



#! /usr/bin/env python3
"""This module contains the Tape class with functions that
do the actual tape imaging.
"""

import os
import sys
import time
import logging
import glob
import _thread as thread
from . import shared

class Tape:
    """Tape class"""
    def __init__(self,
                 dirOut,
                 tapeDevice,
                 initBlockSize,
                 files,
                 prefix,
                 extension,
                 fillBlocks,
                 logFile,
                 SUDO_UID,
                 SUDO_GID):
        """initialise Tape class instance"""

        # Input collected by GUI / CLI
        self.dirOut = dirOut
        self.tapeDevice = tapeDevice
        self.initBlockSize = initBlockSize
        self.files = files
        self.prefix = prefix
        self.extension = extension
        self.fillBlocks = fillBlocks
        self.logFile = logFile
        # Input validation flags
        self.dirOutIsDirectory = False
        self.outputExistsFlag = False
        self.deviceAccessibleFlag = False
        self.dirOutIsWritable = False
        self.blockSizeIsValid = False
        self.filesIsValid = False
        # Miscellaneous attributes
        self.SUDO_UID = SUDO_UID
        self.SUDO_GID = SUDO_GID
        self.tapeDeviceIOError = False
        self.successFlag = True
        self.endOfTape = False
        self.extractFile = False
        self.file = 1
        self.filesList = []
        self.blockSize = 0

    def validateInput(self):
        """Validate and pre-process input"""

        # Check if dirOut is a directory
        self.dirOutIsDirectory = os.path.isdir(self.dirOut)

        # Check if glob pattern for dirOut, prefix and extension matches existing files
        if glob.glob(self.dirOut + '/' + self.prefix + '*.' + self.extension):
            self.outputExistsFlag = True

        # Check if dirOut is writable
        self.dirOutIsWritable = os.access(self.dirOut, os.W_OK | os.X_OK)

        # Check if tape device is accessible
        args = ['mt']
        args.append('-f')
        args.append(self.tapeDevice)
        args.append('status')
        mtStatus, mtOut, mtErr = shared.launchSubProcess(args, False)

        if mtStatus == 0:
            self.deviceAccessibleFlag = True

        # Check if initial block size is valid (i.e. a multiple of 512)
        try:
            self.initBlockSize = int(self.initBlockSize)

            noBlocks = (self.initBlockSize/512)

            if not noBlocks.is_integer():
                self.blockSizeIsValid = False
            elif noBlocks == 0:
                self.blockSizeIsValid = False
            else:
                self.blockSizeIsValid = True
        except ValueError:
            self.blockSizeIsValid = False

        # Check if files entry is valid; also split files string
        # to list of integers
        if self.files.strip() == '':
            # Empty string (default): OK
            self.filesIsValid = True
        else:
            try:
                # Each item in list is an integer: OK
                self.filesList = [int(i) for i in self.files.split(',')]
                self.filesIsValid = True
            except ValueError:
                # One or more items are not an integer
                self.filesIsValid = False

        # Convert fillBlocks to Boolean
        self.fillBlocks = bool(self.fillBlocks)

    def processTape(self):
        """Process a tape"""

        # Write some general info to log file
        logging.info('***************************')
        logging.info('*** TAPE EXTRACTION LOG ***')
        logging.info('***************************\n')
        logging.info('*** USER INPUT ***')
        logging.info('dirOut: ' + self.dirOut)
        logging.info('tapeDevice: ' + self.tapeDevice)
        logging.info('initial blockSize: ' + str(self.initBlockSize))
        logging.info('files: ' + self.files)
        logging.info('prefix: ' + self.prefix)
        logging.info('extension: ' + self.extension)
        logging.info('fill blocks: ' + str(self.fillBlocks))

        if self.fillBlocks:
            # dd's conv=sync flag results in padding bytes for each block if block
            # size is too large, so override user-defined value with default
            # if -f flag was used
            self.initBlockSize = 512
            logging.info('Reset initial block size to 512 because -f flag is used')

        # Get tape status, output to log file
        logging.info('*** Getting tape status ***')

        args = ['mt']
        args.append('-f')
        args.append(self.tapeDevice)
        args.append('status')
        mtStatus, mtOut, mtErr = shared.launchSubProcess(args)

        if mtStatus != 0:
            # Abort if tape device is not accessible
            self.tapeDeviceIOError = True
            self.successFlag = False
            logging.critical('Exiting because tape device is not accessible')
            logging.info('Success: ' + str(self.successFlag))
            # Wait 2 seconds to avoid race condition
            time.sleep(2)
            # This triggers a KeyboardInterrupt in the main thread
            thread.interrupt_main()
            sys.exit()

        # Iterate over all files on tape until end is detected
        while not self.endOfTape:
            # Only extract files defined by files parameter
            # (if file parameter is empty all files are extracted)
            if self.file in self.filesList or self.filesList == []:
                self.extractFile = True
            else:
                self.extractFile = False

            # Call file processing function
            self.processFile()

            # Increase file number
            self.file += 1

        # Create checksum file
        logging.info('*** Creating checksum file ***')
        checksumFile = os.path.join(self.dirOut, "checksums.sha512")
        shared.checksumDirectory(self.dirOut, self.extension, checksumFile)

        # Change owner to user (since script is executed as root)
        chOwnSuccess = shared.changeOwner(checksumFile,
                                          int(self.SUDO_UID),
                                          int(self.SUDO_GID))

        if not chOwnSuccess:
            logging.warning('Could not change owner settings for file ' + checksumFile)

        # Rewind and eject the tape
        logging.info('*** Rewinding tape ***')

        args = ['mt']
        args.append('-f')
        args.append(self.tapeDevice)
        args.append('rewind')
        mtStatus, mtOut, mtErr = shared.launchSubProcess(args)

        logging.info('*** Ejecting tape ***')

        args = ['mt']
        args.append('-f')
        args.append(self.tapeDevice)
        args.append('eject')
        mtStatus, mtOut, mtErr = shared.launchSubProcess(args)

        logging.info('Success: ' + str(self.successFlag))

        if self.successFlag:
            logging.info('Tape processed successfully without errors')
        else:
            logging.error('One or more errors occurred while processing tape, \
            check log file for details')

        # Change owner of log file to user (since script is executed as root)
        chOwnSuccess = shared.changeOwner(self.logFile,
                                       int(self.SUDO_UID),
                                       int(self.SUDO_GID))

        if not chOwnSuccess:
            logging.warning('Could not change owner settings for file ' + self.logFile)

        # Wait 2 seconds to avoid race condition
        time.sleep(2)
        # This triggers a KeyboardInterrupt in the main thread
        thread.interrupt_main()

    def processFile(self):
        """Process a file"""

        if self.extractFile:
            # Determine block size for this file
            logging.info('*** Establishing blockSize ***')
            self.findBlockSize()
            logging.info('Block size: ' + str(self.blockSize))

            # Name of output file for this file
            paddingChars = max(10 - len(self.prefix), 0)
            ofName = self.prefix + str(self.file).zfill(paddingChars) + '.' + self.extension
            ofName = os.path.join(self.dirOut, ofName)

            logging.info('*** Extracting file # ' + str(self.file) + ' to file ' + ofName + ' ***')

            args = ['dd']
            args.append('if=' + self.tapeDevice)
            args.append('of='+ ofName)
            args.append('bs=' + str(self.blockSize))

            if self.fillBlocks:
                # Add conv=noerror,sync options to argument list
                args.append('conv=noerror,sync')

            ddStatus, ddOut, ddErr = shared.launchSubProcess(args)

            # Change owner of extracted file to user (since script is executed as root)
            chOwnSuccess = shared.changeOwner(ofName,
                                              int(self.SUDO_UID),
                                              int(self.SUDO_GID))

            if not chOwnSuccess:
                logging.warning('Could not change owner settings for file ' + ofName)

            if ddStatus != 0:
                self.successFlag = False
                logging.error('dd encountered an error while reading the tape')

        else:
            # Fast-forward tape to next file
            logging.info('*** Skipping file # ' + str(self.file) +
                         ', fast-forward to next file ***')

            args = ['mt']
            args.append('-f')
            args.append(self.tapeDevice)
            args.append('fsf')
            args.append('1')
            mtStatus, mtOut, mtErr = shared.launchSubProcess(args, False)

        # Try to position tape 1 record forward; if this fails this means
        # the end of the tape was reached
        args = ['mt']
        args.append('-f')
        args.append(self.tapeDevice)
        args.append('fsr')
        args.append('1')
        mtStatus, mtOut, mtErr = shared.launchSubProcess(args, False)

        if mtStatus == 0:
            # Another file exists. Position tape one record backward
            args = ['mt']
            args.append('-f')
            args.append(self.tapeDevice)
            args.append('bsr')
            args.append('1')
            mtStatus, mtOut, mtErr = shared.launchSubProcess(args, False)
        else:
            # No further files, end of tape reached
            logging.info('*** Reached end of tape ***')
            self.endOfTape = True

    def findBlockSize(self):
        """Find block size, starting from blockSizeInit"""

        # Set blockSize to initBlockSize
        self.blockSize = self.initBlockSize
        # Flag that indicates block size was found
        blockSizeFound = False

        while not blockSizeFound:
            # Try reading 1 block from tape
            logging.info('*** Guessing block size for file # ' +
                         str(self.file)  + ', trial value ' +
                         str(self.blockSize) + ' ***')

            args = ['dd']
            args.append('if=' + self.tapeDevice)
            args.append('of=/dev/null')
            args.append('bs=' + str(self.blockSize))
            args.append('count=1')
            ddStatus, ddOut, ddErr = shared.launchSubProcess(args, False)

            # Position tape 1 record backward (i.e. to the start of this file)
            args = ['mt']
            args.append('-f')
            args.append(self.tapeDevice)
            args.append('bsr')
            args.append('1')
            mtStatus, mtOut, mtErr = shared.launchSubProcess(args, False)

            if ddStatus == 0:
                # Block size found
                blockSizeFound = True
            else:
                # Try again with larger block size
                self.blockSize += 512



#! /usr/bin/env python3
"""
Tapeimgr, automated reading of tape
Command-line interface

Author: Johan van der Knijff
Research department,  KB / National Library of the Netherlands
"""

import sys
import os
import io
import json
import logging
import argparse
from .tape import Tape
from . import config


class tapeimgrCLI:

    """This class defines the command line interface + associated functions
    for associated actions
    """

    def __init__(self):
        """Initiate class"""
        # Create parser
        # Use formatter to control  column width
        # (see https://stackoverflow.com/a/52606755/1209004 and
        # and https://stackoverflow.com/a/5464440/1209004)
        formatter = lambda prog: argparse.HelpFormatter(prog, max_help_position=50)
        self.parser = argparse.ArgumentParser(description='Read contents of tape. Each file'
                                              ' on the tape is stored as a separate file.',
                                              formatter_class=formatter)
        self.SUDO_USER = ''
        self.SUDO_UID = ''
        self.SUDO_GID = ''
        self.dirOut = ''
        self.logFileName = ''
        self.tapeDevice = ''
        self.initBlockSize = ''
        self.initBlockSizeDefault = ''
        self.files = ''
        self.logFile = ''
        self.prefix = ''
        self.extension = ''
        self.fillBlocks = False
        self.configFile = os.path.normpath('/etc/tapeimgr/tapeimgr.json')
        # Flag that is True if configuration file was read withourt errors, and False if not
        self.configSuccess = True
        # Logging stuff
        self.logger = logging.getLogger()
        # Add stream handler that directs logging output to stdout
        self.consoleHandler = logging.StreamHandler(sys.stdout)

    def parseCommandLine(self):
        """Parse command line"""

        self.parser.add_argument('dirOut',
                                 action='store',
                                 type=str,
                                 help='output directory')
        self.parser.add_argument('--version', '-v',
                                 action='version',
                                 version=config.version)
        self.parser.add_argument('--fill', '-f',
                                 action='store_true',
                                 dest='fillBlocks',
                                 default=self.fillBlocks,
                                 help='fill blocks that give read errors with null bytes')
        self.parser.add_argument('--device', '-d',
                                 action='store',
                                 type=str,
                                 help='non-rewind tape device',
                                 dest='device',
                                 default=self.tapeDevice)
        self.parser.add_argument('--blocksize', '-b',
                                 action='store',
                                 type=str,
                                 help='initial block size (must be a multiple of 512)',
                                 dest='size',
                                 default=self.initBlockSize)
        self.parser.add_argument('--files', '-s',
                                 action='store',
                                 type=str,
                                 help='comma-separated list of files to extract',
                                 dest='files',
                                 default=self.files)
        self.parser.add_argument('--prefix', '-p',
                                 action='store',
                                 type=str,
                                 help='output prefix',
                                 dest='pref',
                                 default=self.prefix)
        self.parser.add_argument('--extension', '-e',
                                 action='store',
                                 type=str,
                                 help='output file extension',
                                 dest='ext',
                                 default=self.extension)

        # Parse arguments
        args = self.parser.parse_args()
        self.dirOut = args.dirOut
        self.fillBlocks = args.fillBlocks
        self.tapeDevice = args.device
        self.initBlockSize = args.size
        self.files = args.files
        self.prefix = args.pref
        self.extension = args.ext

    def getConfiguration(self):
        """read configuration file and set variables accordingly"""
        if not os.path.isfile(self.configFile):
            self.configSuccess = False

        # Read config file to dictionary
        try:
            with io.open(self.configFile, 'r', encoding='utf-8') as f:
                configDict = json.load(f)
        except:
            self.configSuccess = False
        
        if self.configSuccess:
            # Update class variables
            try:
                self.SUDO_USER = configDict['SUDO_USER']
                self.SUDO_UID = configDict['SUDO_UID']
                self.SUDO_GID = configDict['SUDO_GID']
                self.files = configDict['files']
                self.logFileName = configDict['logFileName']
                self.tapeDevice = configDict['tapeDevice']
                self.initBlockSize = configDict['initBlockSize']
                self.initBlockSizeDefault = self.initBlockSize
                self.prefix = configDict['prefix']
                self.extension = configDict['extension']
                self.fillBlocks = bool(configDict['fillBlocks'])
            except KeyError:
                self.configSuccess = False

            try:
                # If executed as root, return normal user's home directory
                self.dirOut = os.path.normpath('/home/' + self.SUDO_USER)
            except TypeError:
                # SUDO_USER doesn't exist if not executed as root
                self.dirOut = os.path.expanduser("~")

    def process(self):
        """fetch and validate entered input, and start processing"""

        # Read configuration file
        self.getConfiguration()

        if not self.configSuccess:
            msg = ("Error reading configuration file! \n" +
                   "Run 'sudo tapeimgr-config' to fix this.")
            errorExit(msg)

        # Parse command line arguments
        self.parseCommandLine()

        # Set logFile
        self.logFile = os.path.join(self.dirOut, self.logFileName)

        # Create tape instance
        self.tape = Tape(self.dirOut,
                         self.tapeDevice,
                         self.initBlockSize,
                         self.files,
                         self.prefix,
                         self.extension,
                         self.fillBlocks,
                         self.logFile,
                         self.SUDO_UID,
                         self.SUDO_GID)

        # Validate input
        self.tape.validateInput()

        # Show error message and exit if any parameters didn't pass validation
        if not self.tape.dirOutIsDirectory:
            msg = ("Output directory '" + self.dirOut + "' doesn't exist!")
            errorExit(msg)

        if not self.tape.dirOutIsWritable:
            msg = ("Cannot write to directory '" + self.dirOut + "'!")
            errorExit(msg)

        if not self.tape.deviceAccessibleFlag:
            msg = ('Tape device is not accessible!')
            errorExit(msg)

        if not self.tape.blockSizeIsValid:
            msg = ("--blocksize '" + str(self.initBlockSize) + "' not valid!")
            errorExit(msg)

        if not self.tape.filesIsValid:
            msg = ('--files value not valid, must be a comma-delimited\n'
                   '    string of integer numbers, or empty!')
            errorExit(msg)

        # Ask confirmation if output files exist already
        if self.tape.outputExistsFlag:
            msg = ('WARNING: writing to ' + self.dirOut + ' will overwrite existing files!\n'
                   'do you really want to proceed? (enter Y to proceed, or N to cancel): ')
            continueResponse = input(msg)

            if continueResponse.upper() == 'N':
                msg = ('Operation cancelled')
                sys.stderr.write(msg)
                sys.exit(1)

        # Start logger
        self.setupLogger()

        # Process the tape
        self.tape.processTape()

    def setupLogger(self):
        """Set up logger configuration"""

        # Basic configuration
        logging.basicConfig(filename=self.logFile,
                            level=logging.INFO,
                            format='%(asctime)s - %(levelname)s - %(message)s')

        # Console handler configuration
        self.consoleHandler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(levelname)s - %(message)s')
        self.consoleHandler.setFormatter(formatter)
        self.logger.addHandler(self.consoleHandler)


def printInfo(msg):
    """Print info to stderr"""
    msgString = ('INFO: ' + msg + '\n')
    sys.stderr.write(msgString)

def printWarning(msg):
    """Print warning to stderr"""
    msgString = ('WARNING: ' + msg + '\n')
    sys.stderr.write(msgString)

def errorExit(msg):
    """Print error to stderr and exit"""
    msgString = ('ERROR: ' + msg + '\n')
    sys.stderr.write(msgString)
    sys.exit(1)


def main():
    """Main command line application"""

    # Create tapeImgrCLI instance
    myCLI = tapeimgrCLI()
    # Start main processing function
    try:
        myCLI.process()
    except KeyboardInterrupt:
        if myCLI.tape.tapeDeviceIOError:
            # Tape device not accessible
            msg = ('Cannot access tape device ' + myCLI.tape.tapeDevice +
                   '. Check that device exists, and that tapeimgr is run as root')
            errorExit(msg)
        elif myCLI.tape.successFlag:
            # Tape extraction completed with no errors
            msg = ('Tape processed successfully without errors!')
            printInfo(msg)
            sys.exit(0)
        else:
            # Tape extraction resulted in errors
            msg = ('One or more errors occurred while processing tape, '
                   'check log file for details')
            errorExit(msg)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
"""Shared functions module"""

import os
import logging
import glob
import hashlib
import subprocess as sub


def launchSubProcess(args, writeLog=True):
    """Launch subprocess and return exit code, stdout and stderr"""
    try:
        # Execute command line; stdout + stderr redirected to objects
        # 'output' and 'errors'.
        # Setting shell=True avoids console window poppong up with pythonw
        # BUT shell=True is not working with argument lists,
        # see https://stackoverflow.com/a/26417712/1209004
        p = sub.Popen(args, stdout=sub.PIPE, stderr=sub.PIPE, shell=False)
        output, errors = p.communicate()

        # Decode to UTF8
        outputAsString = output.decode('utf-8')
        errorsAsString = errors.decode('utf-8')

        exitStatus = p.returncode

    except Exception:
        # I don't even want to to start thinking how one might end up here ...

        exitStatus = -99
        outputAsString = ""
        errorsAsString = ""

    # Logging
    if writeLog:
        cmdName = args[0]
        logging.info('Command: ' + ' '.join(args))

        if exitStatus == 0:
            logging.info(cmdName + ' status: ' + str(exitStatus))
            logging.info(cmdName + ' stdout:\n' + outputAsString)
            logging.info(cmdName + ' stderr:\n' + errorsAsString)
        else:
            logging.error(cmdName + ' status: ' + str(exitStatus))
            logging.error(cmdName + ' stdout:\n' + outputAsString)
            logging.error(cmdName + ' stderr:\n' + errorsAsString)

    return(exitStatus, outputAsString, errorsAsString)


def generate_file_sha512(fileIn):
    """Generate sha512 hash of file"""

    # fileIn is read in chunks to ensure it will work with (very) large files as well
    # Adapted from: http://stackoverflow.com/a/1131255/1209004

    blocksize = 2**20
    m = hashlib.sha512()
    with open(fileIn, "rb") as f:
        while True:
            buf = f.read(blocksize)
            if not buf:
                break
            m.update(buf)
    return m.hexdigest()


def checksumDirectory(directory, extension, checksumFile):
    """Calculate checksums for all files in directory"""

    # All files in directory
    allFiles = glob.glob(directory + "/*." + extension)

    # Dictionary for storing results
    checksums = {}

    for fName in allFiles:
        hashString = generate_file_sha512(fName)
        checksums[fName] = hashString

    # Write checksum file
    try:
        fChecksum = open(checksumFile, "w", encoding="utf-8")
        for fName in checksums:
            lineOut = checksums[fName] + " " + os.path.basename(fName) + '\n'
            fChecksum.write(lineOut)
        fChecksum.close()
        wroteChecksums = True
    except IOError:
        wroteChecksums = False

    return wroteChecksums


def changeOwner(fileRef, uid, gid):
    """Change owner of file"""
    try:
        if os.path.isfile(fileRef):
            os.chown(fileRef, uid, gid)
        successFlag = True
    except:
        successFlag = False

    return successFlag



#! /usr/bin/env python3
"""
Tapeimgr, automated reading of tape
Graphical user interface

Author: Johan van der Knijff
Research department,  KB / National Library of the Netherlands
"""

import sys
import os
import io
import json
import time
import threading
import logging
import queue
import tkinter as tk
from tkinter import filedialog as tkFileDialog
from tkinter import scrolledtext as ScrolledText
from tkinter import messagebox as tkMessageBox
from tkinter import ttk
from .tape import Tape
from . import config


class tapeimgrGUI(tk.Frame):

    """This class defines the graphical user interface + associated functions
    for associated actions
    """

    def __init__(self, parent, *args, **kwargs):
        """Initiate class"""
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.root = parent
        self.SUDO_USER = ''
        self.SUDO_UID = ''
        self.SUDO_GID = ''
        self.dirOut = ''
        self.logFileName = ''
        self.tapeDevice = ''
        self.initBlockSize = ''
        self.initBlockSizeDefault = ''
        self.files = ''
        self.logFile = ''
        self.prefix = ''
        self.extension = ''
        self.fillBlocks = False
        self.configFile = os.path.normpath('/etc/tapeimgr/tapeimgr.json')
        # Flag that is True if configuration file was read withourt errors, and False if not
        self.configSuccess = True
        # Logging stuff
        self.logger = logging.getLogger()
        # Create a logging handler using a queue
        self.log_queue = queue.Queue(-1)
        self.queue_handler = QueueHandler(self.log_queue)
        self.getConfiguration()
        self.build_gui()

    def on_quit(self):
        """Quit tapeimgr"""
        os._exit(0)

    def getConfiguration(self):
        """read configuration file and set variables accordingly"""
        if not os.path.isfile(self.configFile):
            self.configSuccess = False

        # Read config file to dictionary
        try:
            with io.open(self.configFile, 'r', encoding='utf-8') as f:
                configDict = json.load(f)
        except:
            self.configSuccess = False
        
        if self.configSuccess:
            # Update class variables
            try:
                self.SUDO_USER = configDict['SUDO_USER']
                self.SUDO_UID = configDict['SUDO_UID']
                self.SUDO_GID = configDict['SUDO_GID']
                self.files = configDict['files']
                self.logFileName = configDict['logFileName']
                self.tapeDevice = configDict['tapeDevice']
                self.initBlockSize = configDict['initBlockSize']
                self.initBlockSizeDefault = self.initBlockSize
                self.prefix = configDict['prefix']
                self.extension = configDict['extension']
                self.fillBlocks = bool(configDict['fillBlocks'])
            except KeyError:
                self.configSuccess = False

            try:
                # If executed as root, return normal user's home directory
                self.dirOut = os.path.normpath('/home/' + self.SUDO_USER)
            except TypeError:
                # SUDO_USER doesn't exist if not executed as root
                self.dirOut = os.path.expanduser("~")

    def on_submit(self):
        """fetch and validate entered input, and start processing"""

        # This flag is true if all input validates
        inputValidateFlag = True

        # Fetch entered values (strip any leading / trailing whitespace characters)
        self.tapeDevice = self.tapeDevice_entry.get().strip()
        self.initBlockSize = self.initBlockSize_entry.get().strip()
        self.files = self.files_entry.get().strip()
        self.prefix = self.prefix_entry.get().strip()
        self.extension = self.extension_entry.get().strip()
        self.fillBlocks = self.fBlocks.get()
        self.logFile = os.path.join(self.dirOut, self.logFileName)

        # Create tape instance
        self.tape = Tape(self.dirOut,
                         self.tapeDevice,
                         self.initBlockSize,
                         self.files,
                         self.prefix,
                         self.extension,
                         self.fillBlocks,
                         self.logFile,
                         self.SUDO_UID,
                         self.SUDO_GID)

        # Validate input
        self.tape.validateInput()

        # Show error message for any parameters that didn't pass validation
        if not self.tape.dirOutIsDirectory:
            inputValidateFlag = False
            msg = ("Output directory doesn't exist:\n" + self.dirOut)
            tkMessageBox.showerror("ERROR", msg)

        if not self.tape.dirOutIsWritable:
            inputValidateFlag = False
            msg = ('Cannot write to directory ' + self.dirOut)
            tkMessageBox.showerror("ERROR", msg)

        if not self.tape.deviceAccessibleFlag:
            inputValidateFlag = False
            msg = ('Tape device is not accessible')
            tkMessageBox.showerror("ERROR", msg)

        if not self.tape.blockSizeIsValid:
            inputValidateFlag = False
            msg = ('Block size not valid')
            tkMessageBox.showerror("ERROR", msg)

        if not self.tape.filesIsValid:
            inputValidateFlag = False
            msg = ('Files value not valid\n'
                   '(must be comma-delimited string of integer numbers, or empty)')
            tkMessageBox.showerror("ERROR", msg)

        # Ask confirmation if output files exist already
        outDirConfirmFlag = True
        if self.tape.outputExistsFlag:
            msg = ('writing to ' + self.dirOut + ' will overwrite existing files!\n'
                   'press OK to continue, otherwise press Cancel ')
            outDirConfirmFlag = tkMessageBox.askokcancel("Overwrite files?", msg)

        if inputValidateFlag and outDirConfirmFlag:

            # Start logger
            successLogger = True
            try:
                self.setupLogger()
                # Start polling log messages from the queue
                self.after(100, self.poll_log_queue)
            except OSError:
                # Something went wrong while trying to write to lof file
                msg = ('error trying to write log file to ' + self.logFile)
                tkMessageBox.showerror("ERROR", msg)
                successLogger = False

            if successLogger:
                # Disable start and exit buttons
                self.start_button.config(state='disabled')
                self.quit_button.config(state='disabled')

                # Launch tape processing function as subprocess
                t1 = threading.Thread(target=self.tape.processTape)
                t1.start()


    def selectOutputDirectory(self, event=None):
        """Select output directory"""
        dirInit = self.dirOut
        self.dirOut = tkFileDialog.askdirectory(initialdir=dirInit)
        self.outDirLabel['text'] = self.dirOut

    def decreaseBlocksize(self):
        """Decrease value of initBlockSize"""
        try:
            blockSizeOld = int(self.initBlockSize_entry.get().strip())
        except ValueError:
            # Reset if user manually entered something weird
            blockSizeOld = int(self.initBlockSizeDefault)
        blockSizeNew = max(blockSizeOld - 512, 512)
        self.initBlockSize_entry.delete(0, tk.END)
        self.initBlockSize_entry.insert(tk.END, str(blockSizeNew))

    def increaseBlocksize(self):
        """Increase value of initBlockSize"""
        try:
            blockSizeOld = int(self.initBlockSize_entry.get().strip())
        except ValueError:
            # Reset if user manually entered something weird
            blockSizeOld = int(self.initBlockSizeDefault)
        blockSizeNew = blockSizeOld + 512
        self.initBlockSize_entry.delete(0, tk.END)
        self.initBlockSize_entry.insert(tk.END, str(blockSizeNew))

    def build_gui(self):
        """Build the GUI"""

        self.root.title('tapeimgr v.' + config.version)
        self.root.option_add('*tearOff', 'FALSE')
        self.grid(column=0, row=0, sticky='w')
        self.grid_columnconfigure(0, weight=0, pad=0)
        self.grid_columnconfigure(1, weight=0, pad=0)
        self.grid_columnconfigure(2, weight=0, pad=0)
        self.grid_columnconfigure(3, weight=0, pad=0)

        # Entry elements
        ttk.Separator(self, orient='horizontal').grid(column=0, row=0, columnspan=4, sticky='ew')
        # Output Directory
        self.outDirButton_entry = tk.Button(self,
                                            text='Select Output Directory',
                                            command=self.selectOutputDirectory,
                                            width=20)
        self.outDirButton_entry.grid(column=0, row=3, sticky='w')
        self.outDirLabel = tk.Label(self, text=self.dirOut)
        self.outDirLabel.update()
        self.outDirLabel.grid(column=1, row=3, sticky='w')

        ttk.Separator(self, orient='horizontal').grid(column=0, row=5, columnspan=4, sticky='ew')

        # Tape Device
        tk.Label(self, text='Tape Device').grid(column=0, row=6, sticky='w')
        self.tapeDevice_entry = tk.Entry(self, width=20)
        self.tapeDevice_entry['background'] = 'white'
        self.tapeDevice_entry.insert(tk.END, self.tapeDevice)
        self.tapeDevice_entry.grid(column=1, row=6, sticky='w')

        # Initial Block Size
        tk.Label(self, text='Initial Block Size').grid(column=0, row=7, sticky='w')
        self.initBlockSize_entry = tk.Entry(self, width=20)
        self.initBlockSize_entry['background'] = 'white'
        self.initBlockSize_entry.insert(tk.END, self.initBlockSize)
        self.initBlockSize_entry.grid(column=1, row=7, sticky='w')
        self.decreaseBSButton = tk.Button(self, text='-', command=self.decreaseBlocksize, width=1)
        self.decreaseBSButton.grid(column=2, row=7, sticky='e')
        self.increaseBSButton = tk.Button(self, text='+', command=self.increaseBlocksize, width=1)
        self.increaseBSButton.grid(column=3, row=7, sticky='w')

        # Files
        tk.Label(self, text='Files (comma-separated list)').grid(column=0, row=8, sticky='w')
        self.files_entry = tk.Entry(self, width=20)
        self.files_entry['background'] = 'white'
        self.files_entry.grid(column=1, row=8, sticky='w')

        # Prefix
        tk.Label(self, text='Prefix').grid(column=0, row=9, sticky='w')
        self.prefix_entry = tk.Entry(self, width=20)
        self.prefix_entry['background'] = 'white'
        self.prefix_entry.insert(tk.END, self.prefix)
        self.prefix_entry.grid(column=1, row=9, sticky='w')

        # Extension
        tk.Label(self, text='Extension').grid(column=0, row=10, sticky='w')
        self.extension_entry = tk.Entry(self, width=20)
        self.extension_entry['background'] = 'white'
        self.extension_entry.insert(tk.END, self.extension)
        self.extension_entry.grid(column=1, row=10, sticky='w')

        # Fill failed blocks
        tk.Label(self, text='Fill failed blocks').grid(column=0, row=11, sticky='w')
        self.fBlocks = tk.IntVar()
        self.fillblocks_entry = tk.Checkbutton(self, variable=self.fBlocks)
        self.fillblocks_entry.grid(column=1, row=11, sticky='w')

        ttk.Separator(self, orient='horizontal').grid(column=0, row=12, columnspan=4, sticky='ew')

        self.start_button = tk.Button(self,
                                      text='Start',
                                      width=10,
                                      underline=0,
                                      command=self.on_submit)
        self.start_button.grid(column=1, row=13, sticky='e')

        self.quit_button = tk.Button(self,
                                     text='Exit',
                                     width=10,
                                     underline=0,
                                     command=self.on_quit)
        self.quit_button.grid(column=2, row=13, sticky='w', columnspan=2)

        ttk.Separator(self, orient='horizontal').grid(column=0, row=14, columnspan=4, sticky='ew')

        # Add ScrolledText widget to display logging info
        self.st = ScrolledText.ScrolledText(self, state='disabled', height=15)
        self.st.configure(font='TkFixedFont')
        self.st['background'] = 'white'
        self.st.grid(column=0, row=15, sticky='ew', columnspan=4)

        # Define bindings for keyboard shortcuts: buttons
        self.root.bind_all('<Control-Key-d>', self.selectOutputDirectory)
        self.root.bind_all('<Control-Key-s>', self.on_submit)
        self.root.bind_all('<Control-Key-e>', self.on_quit)

        for child in self.winfo_children():
            child.grid_configure(padx=5, pady=5)

        # Display message and exit if config file could not be read
        if not self.configSuccess:
            msg = ("Error reading configuration file! \n" +
                   "Run 'sudo tapeimgr-config' to fix this.")
            errorExit(msg)

    def setupLogger(self):
        """Set up logger configuration"""

        # Basic configuration
        logging.basicConfig(filename=self.logFile,
                            level=logging.INFO,
                            format='%(asctime)s - %(levelname)s - %(message)s')

        # Add the handler to logger
        self.logger = logging.getLogger()

        # This sets the console output format (slightly different from basicConfig!)
        formatter = logging.Formatter('%(levelname)s: %(message)s')
        self.queue_handler.setFormatter(formatter)
        self.logger.addHandler(self.queue_handler)

    def display(self, record):
        """Display log record in scrolledText widget"""
        msg = self.queue_handler.format(record)
        self.st.configure(state='normal')
        self.st.insert(tk.END, msg + '\n', record.levelname)
        self.st.configure(state='disabled')
        # Autoscroll to the bottom
        self.st.yview(tk.END)

    def poll_log_queue(self):
        """Check every 100ms if there is a new message in the queue to display"""
        while True:
            try:
                record = self.log_queue.get(block=False)
            except queue.Empty:
                break
            else:
                self.display(record)
        self.after(100, self.poll_log_queue)


class QueueHandler(logging.Handler):
    """Class to send logging records to a queue

    It can be used from different threads
    The ConsoleUi class polls this queue to display records in a ScrolledText widget
    Taken from https://github.com/beenje/tkinter-logging-text-widget/blob/master/main.py
    """

    def __init__(self, log_queue):
        super().__init__()
        self.log_queue = log_queue

    def emit(self, record):
        self.log_queue.put(record)


def checkDirExists(dirIn):
    """Check if directory exists and exit if not"""
    if not os.path.isdir(dirIn):
        msg = ('directory ' + dirIn + ' does not exist!')
        tkMessageBox.showerror("Error", msg)
        sys.exit(1)


def errorExit(error):
    """Show error message in messagebox and then exit after user presses OK"""
    tkMessageBox.showerror("Error", error)
    os._exit(1)


def main():
    """Main function"""

    packageDir = os.path.dirname(os.path.abspath(__file__))
    root = tk.Tk()
    root.iconphoto(True, tk.PhotoImage(file=os.path.join(packageDir, 'icons', 'tapeimgr.png')))
    myGUI = tapeimgrGUI(root)
    # This ensures application quits normally if user closes window
    root.protocol('WM_DELETE_WINDOW', myGUI.on_quit)

    while True:
        try:
            root.update_idletasks()
            root.update()
            time.sleep(0.1)
        except KeyboardInterrupt:
            if myGUI.tape.tapeDeviceIOError:
                # Tape device not accessible
                msg = ('Cannot access tape device ' + myGUI.tape.tapeDevice +
                       '. Check that device exits, and that tapeimgr is run as root')
                errorExit(msg)
            elif myGUI.tape.successFlag:
                # Tape extraction completed with no errors
                msg = ('Tape processed successfully without errors')
                tkMessageBox.showinfo("Success", msg)
            else:
                # Tape extraction resulted in errors
                msg = ('One or more errors occurred while processing tape, '
                       'check log file for details')
                tkMessageBox.showwarning("Errors occurred", msg)

            # Restart the program
            python = sys.executable
            os.execl(python, python, * sys.argv)
        except:
            # Unexpected error
            msg = 'An unexpected error occurred'
            errorExit(msg)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Setup script for tapeimgr"""

import codecs
import os
import re
from setuptools import setup, find_packages


def read(*parts):
    """Read file and return contents"""
    path = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(path, encoding='utf-8') as fobj:
        return fobj.read()


def find_version(*file_paths):
    """Return version number from main module"""
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


INSTALL_REQUIRES = [
    'setuptools'
]
PYTHON_REQUIRES = '>=3.2'

setup(name='tapeimgr',
      packages=find_packages(),
      version=find_version('tapeimgr', 'tapeimgr.py'),
      license='Apache License 2.0',
      install_requires=INSTALL_REQUIRES,
      python_requires=PYTHON_REQUIRES,
      platforms=['linux'],
      description='Tape imager',
      long_description='Tape imager/reader; wraps around Unix dd and mt tools',
      author='Johan van der Knijff',
      author_email='johan.vanderknijff@kb.nl',
      maintainer='Johan van der Knijff',
      maintainer_email='johan.vanderknijff@kb.nl',
      url='https://github.com/KBNLresearch/tapeimgr',
      download_url=('https://github.com/KBNLresearch/tapeimgr/archive/' +
                    find_version('tapeimgr', 'tapeimgr.py') + '.tar.gz'),
      package_data={'tapeimgr': ['*.*', 'icons/*', 'pkexec/*']},
      zip_safe=False,
      entry_points={'gui_scripts': [
          'tapeimgr = tapeimgr.tapeimgr:main'],
                    'console_scripts': [
                        'tapeimgr = tapeimgr.tapeimgr:main',
                        'tapeimgr-config = tapeimgr.configure:main']},
      classifiers=[
          'Programming Language :: Python :: 3',]
     )
